!+
! Module: compare_realcols
!
! Author: R. Gastaud  - SAp  CEA Saclay
!          
! Description: test if one colummn deviates from a reference
! by less than an absolute or relative tolerance
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module compare_realcols_mod
  
  use emutils_mod
  use param  
  use dal
  use errorhandling

  implicit none

  private
  
! Names of the task parameters

  character(len=*), parameter :: outfile    = "table", &
                                 reffile    = "reftable", &
                                 colname    = "colnames", &
                                 abstol     = "abstol", &
                                 reltol     = "reltol", &
                                 oper       = "operation"

! Routine name

  character(len=*), parameter :: taskname = "compare_realcols"

contains 

  !---------------------------------------------------------------------
  subroutine compare_realcols
  !---------------------------------------------------------------------

    character(len=256)  :: events_out_file, events_ref_file
    character(len=132), allocatable, dimension(:) :: col_name
    character(len=8)    :: toltype
    real(kind=double)   :: tolabs, tolrel

    type(DataSetT)  :: set_out, set_ref
    type(TableT)    :: table_out, table_ref
    real(kind=double), dimension(:), allocatable :: difference, dataout, dataref
    logical,           dimension(:), allocatable :: whereabs
    integer,           dimension(1)              :: ibad
    integer         :: nevents, nevents_out, numcols, icol

! Read parameter values

    events_out_file = stringParameter(outfile)
    events_ref_file = stringParameter(reffile)
    numcols = sizeListParam(colname)
    if (numcols < 1)  return
    allocate(col_name(numcols))
    call readListParam(colname, col_name, numcols)
    tolabs = realParameter(abstol)
    tolrel = realParameter(reltol)
    toltype = changeCase(stringParameter(oper),2)

    write(errstr,*)  "Comparing ", trim(events_out_file), " and ",  &
                     trim(events_ref_file), " at tolerance absolute",  &
                     tolabs, " ", trim(toltype), " relative", tolrel
    call message(AppMsg,Sparse)

! Open tables

    call testName(events_out_file)
    set_out   = dataSet(events_out_file, READ)
    table_out = table(set_out, events_out_file)
    call testName(events_ref_file)
    set_ref   = dataSet(events_ref_file, READ)
    table_ref = table(set_ref, events_ref_file) 

! Compare sizes

    nevents = numberOfRows(table_ref)
    nevents_out = numberOfRows(table_out)
    if (nevents .ne. nevents_out) then
       write(errstr,*) "Tables ", trim(qualifiedName(table_out)),  &
                       " and ", trim(qualifiedName(table_ref)),  &
                       " do not have the same size", nevents, nevents_out
       call error("compare_realcols01")
    endif

    allocate(dataout(nevents), dataref(nevents))
    allocate(difference(nevents), whereabs(nevents))

! Loop over columns

    do icol = 1, numcols

! Open columns

       call readColDble(table_out, col_name(icol), dataout)
       call readColDble(table_ref, col_name(icol), dataref)

! Distinguish between rows where the absolute and relative tolerances apply

       whereabs = abs(dataref) < tolabs/tolrel
       if (toltype == 'AND')  whereabs = .not. whereabs

! Build absolute and relative differences so that the same test applies

       where (whereabs)
          difference = abs(dataout-dataref)
       elsewhere
          difference = abs(dataout/dataref-1) * (tolabs/tolrel)
       endwhere

! Generate error message if significantly different

       if (maxval(difference) > tolabs) then
          ibad = maxloc(difference)
          write(errstr,*) "Tables ", trim(qualifiedName(table_out)), " and ",  &
                          trim(qualifiedName(table_ref)), " are different ",  &
                          "for column ", trim(col_name(icol)),  &
                          " Maximum difference is", dataout(ibad(1)),  &
                          dataref(ibad(1)), " at row", ibad(1)
          call error("compare_realcols02")
       endif
    enddo

    call release(table_out)
    call release(table_ref)
    call release(set_out)
    call release(set_ref)
    deallocate(col_name, dataout, dataref, difference, whereabs)

  end subroutine compare_realcols

! Read a column into a real(double) array

  subroutine readColDble(table, colName, buffer)

! Inputs
! table  : handle to the table
! colName: name of the column
! Outputs
! buffer : contents of the column

    type(TableT),      intent(in)   :: table
    character(len=*),  intent(in)   :: colName
    real(kind=double), dimension(:), intent(out) :: buffer

    type(ColumnT) :: in_col
    integer       :: coltype

! Possible types of columns

    real(kind=double),   dimension(:), pointer   :: dbledata
    real(kind=single),   dimension(:), pointer   :: sngldata

    in_col = column(table, colName, READ)
    coltype = dataType(in_col)
    select case(coltype)
    case (Real64)
       dbledata => real64Data(in_col)
       buffer = dbledata
    case (Real32)
       sngldata => real32Data(in_col)
       buffer = sngldata
    case default
       write(errstr,*) "input column ", trim(qualifiedName(in_col)),  &
                       " is not of real type ", coltype
       call error("readColDble01")
    end select

    write(errstr,*) "Reading ", trim(qualifiedName(in_col)), ". min data =",  &
                    minval(buffer), ". max data =", maxval(buffer)
    call message(AppMsg,Sparse)

    call release(in_col)

  end subroutine readColDble

! Add ':' if name is that of a file

  subroutine testName(tableName)

! In/out: tablename : Name of the file/table

    character(len=*),  intent(inout)   :: tableName

    integer :: i

    i = index(tableName,':')
    if (i == 0)  tableName = trim(tableName) // ':'

  end subroutine testName

end module compare_realcols_mod
