//-----------------------------------------------------------------------------
//
//				 XMM Science Analysis System (SAS)
//				 (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalF90cxxGlue.cc	--- the Fortran90/C++ interface of the CAL
//
//	DESCRIPTION:
//		This file acts as the interface between the CAL API written in
//		Fortran90 and its implementation in C++. All platform-specifics
//		of the C++/F90 interface are centralized in this file.
//
//	CHANGE HISTORY:


#include "common.h"
#include "CalServer.h"
#include "Celestial.h"
#include "CalF90cxxGlue.h"
#include "CalOalUtility.h"
#include "errstr.h"
#include "f90cxx.h"
#include "Dal.h"
#include "Tracer.h"
#include "Handle.h"
#include "STime.h"

#include "EnergyCorrector.h"

#include "EnergyCombinator.h"
#include "CalDataVector.h"

#include <vector>

using namespace CalOalUtility;


namespace {

static bool checkDope(const Dope1 *dp)
{
	PROC

	if (!dp)
		errstr<<"vector not properly passed"
			  <<Sas::fatal("IllegalPointerValue");

//	the following ugly cast is necessary to detect uninitialized pointers
//	from f90 - they get set to 0xffffffff (32 bit even on Dec with
//	64 bit pointers!)

	char *uninitp = reinterpret_cast<char *>(0xffffffff);

	RETURN dp->addr != 0 && dp->addr != uninitp;
} // checkDope()

template<class type>
inline int8 *int8ptr(type *a)	{ return reinterpret_cast<int8 *>(a); }

}	// namespace



//
//	-------------------( CAL_mosPhaBuild )---------------------------------
//
void CXXFUNC(cal_mosphabuild, CAL_mosPhaBuild)
  (const Dope1 *energye1,
   const Dope1 *energye2,
   const Dope1 *pattern ,
   Dope1 *pha,
   const Dope1 *locbkg ,
   const Dope1 *energye3 ,
   const Dope1 *energye4 ,
   const Dope1 *peripix ,
   const Dope1 *flag)
{
  PROC

    float *r_pha = 0;        

    checkDope(energye1);
    checkDope(energye2);
    checkDope(pattern);
    if (checkDope(pha))
      r_pha = FLOATP(pha->addr);

    unsigned vecSize = energye1->dim[0].extent;
    CalReal32Vector energye1Vec(calServer.real32Vector(
		FLOATP(energye1->addr), vecSize));
    CalInt16Vector energye2Vec(calServer.int16Vector(
		SHORTP(energye2->addr), vecSize));
    CalInt8Vector patternVec(int8ptr(pattern->addr), vecSize);
    CalReal32Vector phaVec(calServer.real32Vector(r_pha, vecSize));

//	instantiate EnergyCorrector object

    EnergyCorrector *ecorr = 0; ecorr = calServer.getAtom(ecorr);

// instantiate EnergyCombinator object

    EnergyCombinator ec(ecorr->e1Coefficients(),
			ecorr->e2Coefficients(),
			ecorr->e3Coefficients(),
			ecorr->e4Coefficients());

//	do the combination

    if (energye3 && energye4 && peripix && flag)
      if (checkDope(energye3) && checkDope(energye4) && 
	  checkDope(peripix)  && checkDope(flag)) {
	CalInt16Vector	energye3Vec(calServer.int16Vector(
		SHORTP(energye3->addr), vecSize));
	CalInt16Vector	energye4Vec(calServer.int16Vector(
		SHORTP(energye4->addr), vecSize));
	CalInt8Vector	peripixVec(int8ptr(peripix->addr), vecSize);
	CalInt32Vector	flagVec(INTP(flag->addr), vecSize);

	if (locbkg)
	  if (checkDope(locbkg)) {
	    CalReal32Vector	locbkgVec(calServer.real32Vector(
		FLOATP(locbkg->addr), vecSize));

	    ec.combine(energye1Vec, energye2Vec, patternVec,
		       phaVec, locbkgVec, energye3Vec, energye4Vec,
		       peripixVec, flagVec);
	  } else
	    errstr<<"locbkg vector is empty"
		  <<Sas::error("EmptyVector");
	else
	  ec.combine(energye1Vec, energye2Vec, patternVec,
		     phaVec, energye3Vec, energye4Vec,
		     peripixVec, flagVec);
      } else
	errstr<<"energy3/energy4/peripix/flag vector(s) is/are empty"
	      <<Sas::error("EmptyVector");
    else
      if (locbkg)
	if (checkDope(locbkg)) {
	  CalReal32Vector	locbkgVec(calServer.real32Vector(
		FLOATP(locbkg->addr), vecSize));
	  ec.combine(energye1Vec, energye2Vec, patternVec,
		     phaVec, locbkgVec);
	} else
	  errstr<<"locbkg vector is empty"
		<<Sas::error("EmptyVector");
      else
	ec.combine(energye1Vec, energye2Vec, patternVec, phaVec);

    if (!checkDope(pha)) {

//	setup the dope stuff properly

      vector<unsigned>	oneDim(1); oneDim[0] = vecSize;
      setupDope(*pha, phaVec.ptr(), oneDim);
    }

ENDPROC
} // CAL_mosPhaBuild()
